// Copyright 2018-present Network Optix, Inc. Licensed under MPL 2.0: www.mozilla.org/MPL/2.0/

#include "device_agent.h"

#include <chrono>

#define NX_PRINT_PREFIX (this->logUtils.printPrefix)
#include <nx/kit/debug.h>

#include <nx/sdk/analytics/helpers/object_metadata.h>
#include <nx/sdk/analytics/helpers/object_metadata_packet.h>

#include "device_agent_manifest.h"
#include "object_generators.h"

namespace nx {
namespace vms_server_plugins {
namespace analytics {
namespace stub {
namespace taxonomy_features {

using namespace nx::sdk;
using namespace nx::sdk::analytics;

static constexpr int kTrackLength = 200;

static Rect generateBoundingBox(int frameIndex, int trackIndex, int trackCount)
{
    Rect boundingBox;
    boundingBox.width = 1.0F / trackCount;
    boundingBox.height = boundingBox.width;
    boundingBox.x = 1.0F / trackCount * trackIndex;
    boundingBox.y = 1.0F - boundingBox.height - (1.0F / kTrackLength) * (frameIndex % kTrackLength);

    return boundingBox;
}

Ptr<IMetadataPacket> DeviceAgent::generateObjectMetadataPacket(int64_t frameTimestampUs)
{
    auto metadataPacket = makePtr<ObjectMetadataPacket>();
    metadataPacket->setTimestampUs(frameTimestampUs);

    std::vector<Ptr<ObjectMetadata>> objects{
        generateInstanceOfBaseObjectType(),
        generateInstanceOfDerivedObjectType(),
        generateInstanceOfDerivedObjectTypeWithOmittedAttributes(),
        generateInstanceOfHiddenDerivedObjectType(),
        generateInstanceOfHiddenDerivedObjectTypeWithOwnAttributes(),
        generateInstanceOfDerivedObjectTypeWithUnsupportedBase(),
        generateInstanceOfObjectTypeWithNumericAttibutes(),
        generateInstanceOfObjectTypeWithBooleanAttibutes(),
        generateInstanceOfObjectTypeWithIcon(),
        generateInstanceOfObjectTypeInheritedFromBaseLibraryType(),
        generateInstanceOfObjectTypeUsingBaseLibraryEnumType(),
        generateInstanceOfObjectTypeUsingBaseLibraryColorType(),
        generateInstanceOfObjectTypeUsingBaseLibraryObjectType(),
        generateInstanceOfOfBaseLibraryObjectType(),
        generateInstanceOfObjectTypeDeclaredInEngineManifest(),
        generateInstanceOfLiveOnlyObjectType(),
        generateInstanceOfNonIndexableObjectType(),
    };

    for (int i = 0; i < (int) objects.size(); ++i)
    {
        objects[i]->setBoundingBox(generateBoundingBox(m_frameIndex, i, (int) objects.size()));
        objects[i]->setTrackId(trackIdByTrackIndex(i));

        metadataPacket->addItem(objects[i].get());
    }

    return metadataPacket;
}

DeviceAgent::DeviceAgent(Engine* engine, const nx::sdk::IDeviceInfo* deviceInfo):
    ConsumingDeviceAgent(deviceInfo, /*enableOutput*/ true, engine->plugin()->instanceId()),
    m_engine(engine)
{
}

DeviceAgent::~DeviceAgent()
{
}

std::string DeviceAgent::manifestString() const
{
    return kDeviceAgentManifest;
}

bool DeviceAgent::pushCompressedVideoFrame(const ICompressedVideoPacket* videoFrame)
{
    ++m_frameIndex;
    if ((m_frameIndex % kTrackLength) == 0)
        m_trackIds.clear();

    Ptr<IMetadataPacket> objectMetadataPacket = generateObjectMetadataPacket(
        videoFrame->timestampUs());

    pushMetadataPacket(objectMetadataPacket.releasePtr());

    return true;
}

void DeviceAgent::doSetNeededMetadataTypes(
    nx::sdk::Result<void>* /*outValue*/,
    const nx::sdk::analytics::IMetadataTypes* /*neededMetadataTypes*/)
{
}

Uuid DeviceAgent::trackIdByTrackIndex(int trackIndex)
{
    while (trackIndex >= (int) m_trackIds.size())
        m_trackIds.push_back(UuidHelper::randomUuid());

    return m_trackIds[trackIndex];
}

} // namespace taxonomy_features
} // namespace stub
} // namespace analytics
} // namespace vms_server_plugins
} // namespace nx
